import time
from aiohttp import web
from typing import Optional
from asyncworker.rabbitmq.message import RabbitMQMessage
from barterdude import BarterDude
from barterdude.hooks import HttpHook
from barterdude.hooks.metrics.prometheus.definitions import Definitions
from barterdude.hooks.metrics.prometheus.metrics import Metrics
try:
    from prometheus_client import (
        CollectorRegistry,
        generate_latest,
        CONTENT_TYPE_LATEST,
    )
except ImportError:  # pragma: no cover
    raise ImportError("""
    Please install extra dependency with:
        `pip install barterdude[prometheus]`
    """)


class Prometheus(HttpHook):

    def __init__(
        self,
        barterdude: BarterDude,
        labels: dict,
        path: str = "/metrics",
        registry: CollectorRegistry = None
    ):
        self.__registry = registry or CollectorRegistry()
        self.__labels = labels
        self.__metrics = Metrics(self.__registry)
        self.__definitions = Definitions(
            self.__registry, self.__metrics, list(labels.keys())
        )
        self._msg_start = {}
        self.__definitions.save_metrics()
        super(Prometheus, self).__init__(barterdude, path)

    @property
    def metrics(self):
        return self.__metrics

    async def before_consume(self, message: RabbitMQMessage):
        hash_message = id(message)
        self.metrics[self.__definitions.BEFORE_CONSUME].labels(
            **self.__labels
        ).inc()
        self._msg_start[hash_message] = time.time()

    async def _on_complete(self,
                           message: RabbitMQMessage,
                           state: str,
                           error: Optional[Exception] = None):

        final_time = time.time()
        hash_message = id(message)
        labels = self.__labels.copy()
        labels["state"] = state
        labels["error"] = str(type(error)) if (error) else None
        self.metrics[state].labels(**labels).inc()
        self.metrics[self.__definitions.TIME_MEASURE].labels(
            **labels).observe(
                final_time - self._msg_start.pop(hash_message)
        )

    async def on_success(self, message: RabbitMQMessage):
        await self._on_complete(message, self.__definitions.SUCCESS)

    async def on_fail(self, message: RabbitMQMessage, error: Exception):
        await self._on_complete(message, self.__definitions.FAIL, error)

    async def on_connection_fail(self, error: Exception, retries: int):
        self.metrics[self.__definitions.CONNECTION_FAIL].labels(
            **self.__labels
        ).inc()

    async def __call__(self, req: web.Request):
        return web.Response(
            content_type=CONTENT_TYPE_LATEST.split(";")[0],
            charset=CONTENT_TYPE_LATEST.split("=")[-1],
            body=generate_latest(self.__registry),
            status=200,
        )



from asynctest import Mock, TestCase, CoroutineMock, patch, call
from asyncworker import Options, RouteTypes
from barterdude import BarterDude
from barterdude.message import Message
from tests_unit.helpers import load_fixture


class TestBarterDude(TestCase):
    @patch("barterdude.App")
    @patch("barterdude.AMQPConnection")
    def setUp(self, AMQPConnection, App):
        self.monitor = Mock()
        self.monitor.dispatch_before_consume = CoroutineMock()
        self.monitor.dispatch_on_success = CoroutineMock()
        self.monitor.dispatch_on_fail = CoroutineMock()
        self.callback = CoroutineMock()
        self.messages = [Mock(value=i) for i in range(10)]
        self.calls = [call(message) for message in self.messages]

        self.AMQPConnection = AMQPConnection
        self.connection = self.AMQPConnection.return_value
        self.App = App
        self.app = self.App.return_value
        self.app.startup = CoroutineMock()
        self.app.shutdown = CoroutineMock()
        self.decorator = self.app.route.return_value
        self.schema = load_fixture("schema.json")
        self.barterdude = BarterDude()

    def test_should_create_connection(self):
        self.AMQPConnection.assert_called_once_with(  # nosec
            hostname="127.0.0.1",
            username="guest",
            password="guest",
            prefetch=10,
            name="default",
        )
        self.App.assert_called_once_with(connections=[self.connection])

    def test_should_call_route_when_created(self):
        monitor = Mock()
        self.barterdude.consume_amqp(
            ["queue"], monitor=monitor
        )(CoroutineMock())
        self.app.route.assert_called_once_with(
            ["queue"],
            type=RouteTypes.AMQP_RABBITMQ,
            options={
                Options.BULK_SIZE: 10,
                Options.BULK_FLUSH_INTERVAL: 60,
                Options.CONNECTION_FAIL_CALLBACK:
                    monitor.dispatch_on_connection_fail,
            }
        )

    def test_should_call_route_when_adding_endpoint(self):
        hook = Mock()
        self.barterdude.add_endpoint(['/my_route'], ['GET'], hook)
        self.app.route.assert_called_once_with(
            routes=['/my_route'],
            methods=['GET'],
            type=RouteTypes.HTTP
        )
        self.decorator.assert_called_once_with(hook)

    async def test_should_call_callback_for_each_message(self):
        self.barterdude.consume_amqp(["queue"], self.monitor)(self.callback)
        self.decorator.assert_called_once()
        wrapper = self.decorator.call_args[0][0]
        await wrapper(self.messages)
        messages = []
        for message in self.callback.mock_calls:
            self.assertEqual(Message, type(message[1][0]))
            messages.append(message[1][0]._message)
        self.assertListEqual(
            sorted(messages, key=lambda x: x.value),
            sorted(self.messages, key=lambda x: x.value))

    async def test_should_call_reject_when_callback_fail(self):
        self.callback.side_effect = Exception('Boom!')
        self.barterdude.consume_amqp(["queue"], self.monitor)(self.callback)
        wrapper = self.decorator.call_args[0][0]
        await wrapper(self.messages)
        for message in self.messages:
            message.reject.assert_called_once()

    async def test_should_call_monitor_for_each_success_message(self):
        self.barterdude.consume_amqp(["queue"], self.monitor)(self.callback)
        wrapper = self.decorator.call_args[0][0]
        await wrapper(self.messages)
        self.monitor.dispatch_before_consume.assert_has_calls(
            self.calls, any_order=True)
        self.monitor.dispatch_on_success.assert_has_calls(
            self.calls, any_order=True)
        self.monitor.dispatch_on_fail.assert_not_called()

    async def test_should_call_callback_for_valid_message(self):
        self.barterdude.consume_amqp(
            ["queue"], self.monitor, validation_schema=self.schema
        )(self.callback)
        self.decorator.assert_called_once()
        wrapper = self.decorator.call_args[0][0]
        message = Mock(Message)
        message.body = {"key": 'ok'}
        await wrapper([message])
        self.callback.assert_called_once()
        self.assertEqual(
            self.callback.await_args[0][0].body["key"],
            message.body["key"]
        )

    async def test_should_not_call_callback_for_valid_message(self):
        self.barterdude.consume_amqp(
            ["queue"], self.monitor, validation_schema=self.schema
        )(self.callback)
        self.decorator.assert_called_once()
        wrapper = self.decorator.call_args[0][0]
        message = Mock(Message)
        message.body = {"wrong": 'ok'}
        await wrapper([message])
        self.callback.assert_not_called()

    async def test_should_call_monitor_for_each_fail_message(self):
        error = Exception('Boom!')
        self.callback.side_effect = error
        self.barterdude.consume_amqp(["queue"], self.monitor)(self.callback)
        wrapper = self.decorator.call_args[0][0]
        await wrapper(self.messages)
        self.monitor.dispatch_before_consume.assert_has_calls(
            self.calls, any_order=True)
        error_calls = [call(message, error) for message in self.messages]
        self.monitor.dispatch_on_fail.assert_has_calls(
            error_calls, any_order=True)
        self.monitor.dispatch_on_success.assert_not_called()

    async def test_should_call_put_when_publish(self):
        data = Mock()
        self.connection.put = CoroutineMock()
        await self.barterdude.publish_amqp(
            'exchange',
            data,
            vhost="vhost",
            routing_key="routing_key"
        )
        self.connection.put.assert_called_once_with(
            exchange='exchange',
            data=data,
            vhost="vhost",
            routing_key="routing_key",
            properties=None
        )

    async def test_should_call_startup_and_shutdown(self):
        await self.barterdude.startup()
        self.app.startup.assert_called_once_with()
        await self.barterdude.shutdown()
        self.app.shutdown.assert_called_once_with()

    def test_should_call_run(self):
        self.barterdude.run()
        self.app.run.assert_called_once_with()


class TestAppSharedProperties(TestCase):
    def setUp(self):
        self.barterdude = BarterDude()

    def test_setitem_changes_state(self):
        self.barterdude["foo"] = foo = Mock()
        self.assertEqual(foo, self.barterdude["foo"])

    async def test_getitem_returns_internal_state_value(self):
        self.barterdude["foo"] = "bar"
        self.assertEqual("bar", self.barterdude["foo"])

    def test_delitem_changes_state(self):
        self.barterdude["foo"] = foo = Mock()
        self.assertEqual(foo, self.barterdude["foo"])

        del self.barterdude["foo"]
        with self.assertRaises(KeyError):
            self.assertIsNone(self.barterdude["foo"])

    def test_len_returns_state_len(self):
        test_data = {"foo": 1, "bar": 2}
        for k, v in test_data.items():
            self.barterdude[k] = v

        self.assertEqual(
            len(self.barterdude),
            len(dict(self.barterdude._BarterDude__app))
        )

    async def test_iter_iters_through_internal_state_value(self):
        test_data = {"foo": 1, "bar": 2}
        for k, v in test_data.items():
            self.barterdude[k] = v

        state = dict(**self.barterdude)
        self.assertDictContainsSubset(test_data, state)



from barterdude.hooks import BaseHook
from asyncworker.rabbitmq.message import RabbitMQMessage


class ErrorHook(BaseHook):
    async def on_success(self, message: RabbitMQMessage):
        raise NotImplementedError

    async def on_fail(self, message: RabbitMQMessage, error: Exception):
        raise NotImplementedError

    async def before_consume(self, message: RabbitMQMessage):
        raise NotImplementedError

    async def on_connection_fail(self, error: Exception, retries: int):
        raise NotImplementedError



from asynctest import TestCase, MagicMock, Mock, CoroutineMock, patch
from barterdude.monitor import Monitor


class TestMonitor(TestCase):

    def setUp(self):
        self.hook1 = Mock()
        self.hook2 = Mock()
        self.monitor = Monitor(self.hook1, self.hook2)

    async def test_should_call_hooks_before_consume(self):
        self.hook1.before_consume = CoroutineMock()
        self.hook2.before_consume = CoroutineMock()
        await self.monitor.dispatch_before_consume({})
        self.hook1.before_consume.assert_called_once()
        self.hook2.before_consume.assert_called_once()
        self.hook1.before_consume.assert_called_with({})
        self.hook2.before_consume.assert_called_with({})

    async def test_should_call_hooks_on_success(self):
        self.hook1.on_success = CoroutineMock()
        self.hook2.on_success = CoroutineMock()
        await self.monitor.dispatch_on_success({})
        self.hook1.on_success.assert_called_once()
        self.hook2.on_success.assert_called_once()
        self.hook1.on_success.assert_called_with({})
        self.hook2.on_success.assert_called_with({})

    async def test_should_call_hooks_on_fail(self):
        exception = Exception()
        self.hook1.on_fail = CoroutineMock()
        self.hook2.on_fail = CoroutineMock()
        await self.monitor.dispatch_on_fail({}, exception)
        self.hook1.on_fail.assert_called_once()
        self.hook2.on_fail.assert_called_once()
        self.hook1.on_fail.assert_called_with({}, exception)
        self.hook2.on_fail.assert_called_with({}, exception)

    async def test_should_call_hooks_on_connection_fail(self):
        exception = Mock()
        retries = Mock()
        self.hook1.on_connection_fail = CoroutineMock()
        self.hook2.on_connection_fail = CoroutineMock()
        await self.monitor.dispatch_on_connection_fail(exception, retries)
        self.hook1.on_connection_fail.assert_called_once()
        self.hook2.on_connection_fail.assert_called_once()
        self.hook1.on_connection_fail.assert_called_with(exception, retries)
        self.hook2.on_connection_fail.assert_called_with(exception, retries)

    @patch("barterdude.monitor.repr")
    @patch("barterdude.monitor.format_tb")
    async def test_should_log_if_hook_fail(self, format_tb, repr):
        logger = MagicMock()
        exception = Exception()

        async def before_consume(msg):
            raise exception
        self.monitor._logger = logger
        self.hook1.before_consume = before_consume
        self.hook2.before_consume = CoroutineMock()
        await self.monitor.dispatch_before_consume({})
        repr.assert_called_once_with(exception)
        format_tb.assert_called_once_with(exception.__traceback__)
        logger.error.assert_called_once_with({
            "message": f"Error on hook method {before_consume}",
            "exception": repr.return_value,
            "traceback": format_tb.return_value,
        })
        self.hook2.before_consume.assert_called_with({})



import json

from barterdude import BarterDude
from barterdude.hooks import HttpHook
from asyncworker.rabbitmq.message import RabbitMQMessage
from aiohttp import web
from time import time
from collections import deque
from bisect import bisect_left


def _remove_old(instants: deque, old_timestamp: float):
    pos = bisect_left(instants, old_timestamp)
    for i in range(0, pos):
        instants.popleft()
    return len(instants)


def _response(status, body):
    body["status"] = "ok" if status == 200 else "fail"
    return web.Response(status=status, body=json.dumps(body))


class Healthcheck(HttpHook):
    def __init__(
        self,
        barterdude: BarterDude,
        path: str = "/healthcheck",
        success_rate: float = 0.95,
        health_window: float = 60.0,  # seconds
        max_connection_fails: int = 3
    ):
        self.__success_rate = success_rate
        self.__health_window = health_window
        self.__success = deque()
        self.__fail = deque()
        self.__force_fail = False
        self.__connection_fails = 0
        self.__max_connection_fails = max_connection_fails
        super(Healthcheck, self).__init__(barterdude, path)

    def force_fail(self):
        self.__force_fail = True

    async def before_consume(self, message: RabbitMQMessage):
        pass

    async def on_success(self, message: RabbitMQMessage):
        self.__success.append(time())

    async def on_fail(self, message: RabbitMQMessage, error: Exception):
        self.__fail.append(time())

    async def on_connection_fail(self, error: Exception, retries: int):
        self.__connection_fails = retries

    async def __call__(self, req: web.Request):
        if self.__force_fail:
            return _response(500, {
                "message": "Healthcheck fail called manually"
            })

        if self.__connection_fails >= self.__max_connection_fails:
            return _response(500, {
                "message": (
                    "Reached max connection fails "
                    f"({self.__max_connection_fails})"
                )
            })

        old_timestamp = time() - self.__health_window
        success = _remove_old(self.__success, old_timestamp)
        fail = _remove_old(self.__fail, old_timestamp)
        if success == 0 and fail == 0:
            return _response(200, {
                "message": f"No messages in last {self.__health_window}s"
            })

        rate = success / (success + fail)
        return _response(200 if rate >= self.__success_rate else 500, {
            "message":
                f"Success rate: {rate} (expected: {self.__success_rate})",
            "fail": fail,
            "success": success
        })



import logging
from pythonjsonlogger import jsonlogger

from asynctest import TestCase
from barterdude.conf import (
    getLogger, BARTERDUDE_DEFAULT_LOG_NAME, BARTERDUDE_DEFAULT_LOG_LEVEL
)


class TestConf(TestCase):

    def setUp(self):
        self.log_name = BARTERDUDE_DEFAULT_LOG_NAME
        self.log_level = BARTERDUDE_DEFAULT_LOG_LEVEL

    async def test_should_get_log_with_default_configs(self):
        logger = getLogger("test")
        self.assertEqual(
            type(logger.handlers[0]),
            logging.StreamHandler
        )
        self.assertEqual(logger.name, f"{self.log_name}.test")
        self.assertEqual(logger.level, self.log_level)
        self.assertEqual(
            type(logger.handlers[0].formatter),
            jsonlogger.JsonFormatter
        )

    async def test_should_get_log_with_custom_configs(self):
        logger = getLogger("test", logging.DEBUG)
        self.assertEqual(
            type(logger.handlers[0]),
            logging.StreamHandler
        )
        self.assertEqual(logger.name, f"{self.log_name}.test")
        self.assertEqual(logger.level, logging.DEBUG)
        self.assertEqual(
            type(logger.handlers[0].formatter),
            jsonlogger.JsonFormatter
        )

    async def test_should_get_log_with_custom_configs_even_called_after(self):
        logger_first = getLogger("test_log")
        logger = getLogger("test", logging.DEBUG)
        logger_last = getLogger("test_log")
        self.assertEqual(
            type(logger.handlers[0]),
            logging.StreamHandler
        )
        self.assertEqual(logger_first.name, f"{self.log_name}.test_log")
        self.assertEqual(logger_first.level, self.log_level)
        self.assertEqual(logger.name, f"{self.log_name}.test")
        self.assertEqual(logger.level, logging.DEBUG)
        self.assertEqual(logger_last.name, f"{self.log_name}.test_log")
        self.assertEqual(logger_last.level, self.log_level)
        self.assertEqual(
            type(logger.handlers[0].formatter),
            jsonlogger.JsonFormatter
        )
        self.assertEqual(len(logger_first.handlers), 1)
        self.assertEqual(len(logger.handlers), 1)
        self.assertEqual(len(logger_last.handlers), 1)



from asynctest import TestCase, Mock
from asyncworker.rabbitmq.message import RabbitMQMessage
from barterdude.message import (
    Message, MessageValidation, ValidationException)
from tests_unit.helpers import load_fixture


class TestMessage(TestCase):
    def test_should_convert_rabbitmq_message_to_message(self):
        rbmq_message = RabbitMQMessage(1, Mock())
        message = Message(rbmq_message)
        self.assertEqual(message.body, rbmq_message.body)
        self.assertEqual(
            message.properties, rbmq_message._amqp_message._properties)
        self.assertEqual(
            message.queue_name, rbmq_message._amqp_message.queue_name)
        self.assertEqual(message.raw, rbmq_message.serialized_data)

    async def test_should_call_rbmq_methods(self):
        rbmq_message = Mock(RabbitMQMessage)
        message = Message(rbmq_message)
        message.reject()
        rbmq_message.reject.assert_called_once_with(True)
        rbmq_message.reset_mock()
        message.reject(False)
        rbmq_message.reject.assert_called_once_with(False)
        rbmq_message.reset_mock()
        message.accept()
        rbmq_message.accept.assert_called_once_with()
        rbmq_message.reset_mock()
        await message.process_success()
        rbmq_message.process_success.assert_awaited_with()
        rbmq_message.reset_mock()
        await message.process_exception()
        rbmq_message.process_exception.assert_awaited_with()

    def test_should_work_with_dict(self):
        test_message = {"key": "value"}
        rbmq_message = Mock(RabbitMQMessage)
        rbmq_message.body = test_message
        message = Message(rbmq_message)
        self.assertEqual(message.body, test_message)

    def test_should_work_with_no_dict(self):
        test_message = "value"
        rbmq_message = Mock(RabbitMQMessage)
        rbmq_message.body = test_message
        message = Message(rbmq_message)
        self.assertEqual(message.body, test_message)


class TestMessageValidation(TestCase):
    def setUp(self):
        self.schema = load_fixture("schema.json")
        self.rbmq_message = Mock(RabbitMQMessage)

    def test_should_return_without_validation_even_wrong(self):
        self.rbmq_message.body = {"wrong": "key"}
        validation = MessageValidation()
        message = validation(self.rbmq_message)
        validation.validate(message)
        self.assertEqual(self.rbmq_message.body, message.body)

    def test_should_return_message_with_validation(self):
        self.rbmq_message.body = {"key": "ok"}
        validation = MessageValidation(self.schema)
        message = validation(self.rbmq_message)
        validation.validate(message)
        self.assertEqual(self.rbmq_message.body, message.body)

    def test_should_raise_error_with_wrong_message(self):
        self.rbmq_message.body = {"wrong": "key"}
        validation = MessageValidation(self.schema)
        with self.assertRaises(ValidationException):
            validation(self.rbmq_message)
        with self.assertRaises(ValidationException):
            validation.validate(self.rbmq_message)



import json
from traceback import format_tb
from asyncworker.rabbitmq.message import RabbitMQMessage

from barterdude.conf import (
    getLogger, BARTERDUDE_DEFAULT_LOG_LEVEL
)
from barterdude.hooks import BaseHook


class Logging(BaseHook):

    def __init__(
        self, name="hook.logging",
        level=BARTERDUDE_DEFAULT_LOG_LEVEL
    ):
        self._logger = getLogger(name, level)

    @property
    def logger(self):
        return self._logger

    async def before_consume(self, message: RabbitMQMessage):
        self.logger.info({
            "message": "Before consume message",
            "delivery_tag": message._delivery_tag,
            "message_body": json.dumps(message.body),
        })

    async def on_success(self, message: RabbitMQMessage):
        self.logger.info({
            "message": "Successfully consumed message",
            "delivery_tag": message._delivery_tag,
            "message_body": json.dumps(message.body),
        })

    async def on_fail(self, message: RabbitMQMessage, error: Exception):
        self.logger.error({
            "message": "Failed to consume message",
            "delivery_tag": message._delivery_tag,
            "message_body": json.dumps(message.body),
            "exception": repr(error),
            "traceback": format_tb(error.__traceback__),
        })

    async def on_connection_fail(self, error: Exception, retries: int):
        self.logger.error({
            "message": "Failed to connect to the broker",
            "retries": retries,
            "exception": repr(error),
            "traceback": format_tb(error.__traceback__),
        })



import os
import aiohttp
import asyncio

from asynctest import TestCase
from barterdude import BarterDude
from barterdude.monitor import Monitor
from barterdude.hooks.healthcheck import Healthcheck
from barterdude.hooks.logging import Logging
from barterdude.hooks.metrics.prometheus import Prometheus
from barterdude.message import ValidationException
from tests_unit.helpers import load_fixture
from tests_integration.helpers import ErrorHook
from asyncworker.connections import AMQPConnection
from random import choices
from string import ascii_uppercase


class RabbitMQConsumerTest(TestCase):
    use_default_loop = True

    async def setUp(self):
        self.input_queue = "test"
        self.output_exchange = "test_exchange"
        self.output_queue = "test_output"
        self.rabbitmq_host = os.environ.get("RABBITMQ_HOST", "127.0.0.1")
        self.barterdude_host = os.environ.get("BARTERDUDE_HOST", "127.0.0.1")

        self.connection = AMQPConnection(  # nosec
            name="barterdude-test",
            hostname=self.rabbitmq_host,
            username="guest",
            password="guest",
            prefetch=1
        )
        self.queue_manager = self.connection["/"]
        await self.queue_manager.connection._connect()
        await self.queue_manager.connection.channel.queue_declare(
            self.input_queue
        )
        await self.queue_manager.connection.channel.exchange_declare(
            self.output_exchange, "direct"
        )
        await self.queue_manager.connection.channel.queue_declare(
            self.output_queue
        )
        await self.queue_manager.connection.channel.queue_bind(
            self.output_queue, self.output_exchange, ""
        )

        self.messages = []
        for i in range(10):
            message = {"key": "".join(choices(ascii_uppercase, k=16))}
            self.messages.append(message)

        self.schema = load_fixture("schema.json")

        self.app = BarterDude(hostname=self.rabbitmq_host)

    async def tearDown(self):
        await self.app.shutdown()
        await self.queue_manager.connection.channel.queue_delete(
            self.input_queue
        )
        await self.queue_manager.connection.channel.queue_delete(
            self.output_queue
        )
        await self.queue_manager.connection.channel.exchange_delete(
            self.output_exchange
        )
        await self.queue_manager.connection.close()

    async def send_all_messages(self):
        futures = []
        for message in self.messages:
            futures.append(self.queue_manager.put(
                routing_key=self.input_queue,
                data=message
            ))
        await asyncio.gather(*futures)

    async def test_process_messages_successfully(self):
        received_messages = set()

        @self.app.consume_amqp([self.input_queue], coroutines=1)
        async def handler(message):
            nonlocal received_messages
            received_messages.add(message.body["key"])

        await self.app.startup()
        await self.send_all_messages()
        await asyncio.sleep(1)

        for message in self.messages:
            self.assertTrue(message["key"] in received_messages)

    async def test_process_messages_successfully_with_validation(self):
        received_messages = set()

        @self.app.consume_amqp(
            [self.input_queue], coroutines=1, validation_schema=self.schema)
        async def handler(message):
            nonlocal received_messages
            received_messages.add(message.body["key"])

        await self.app.startup()
        await self.send_all_messages()
        await asyncio.sleep(1)

        for message in self.messages:
            self.assertTrue(message["key"] in received_messages)

    async def test_process_messages_successfully_even_with_crashed_hook(self):
        received_messages = set()

        monitor = Monitor(ErrorHook())

        @self.app.consume_amqp(
            [self.input_queue], coroutines=1, monitor=monitor)
        async def handler(message):
            nonlocal received_messages
            received_messages.add(message.body["key"])

        await self.app.startup()
        await self.send_all_messages()
        await asyncio.sleep(1)

        for message in self.messages:
            self.assertTrue(message["key"] in received_messages)

    async def test_process_one_message_and_publish(self):
        @self.app.consume_amqp([self.input_queue], coroutines=1)
        async def forward(message):
            await self.app.publish_amqp(
                self.output_exchange,
                self.messages[1]
            )

        received_message = None

        @self.app.consume_amqp([self.output_queue], coroutines=1)
        async def handler(message):
            nonlocal received_message
            received_message = message.body

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)
        self.assertEquals(received_message, self.messages[1])

    async def test_process_message_requeue_with_requeue(self):
        handler_called = 0

        @self.app.consume_amqp([self.input_queue], coroutines=1)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise Exception()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)
        self.assertEquals(handler_called, 2)

    async def test_process_message_reject_with_requeue(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], requeue_on_fail=True,
            bulk_flush_interval=1, coroutines=1)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise Exception()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(2)
        self.assertEquals(handler_called, 2)

    async def test_process_message_reject_by_validation_with_requeue(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], requeue_on_validation_fail=True,
            bulk_flush_interval=1, coroutines=1)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise ValidationException()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(2)
        self.assertEquals(handler_called, 2)

    async def test_process_message_reject_without_requeue(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], requeue_on_fail=False,
            bulk_flush_interval=1, coroutines=1)
        async def handler(message):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise Exception()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(2)
        self.assertEquals(handler_called, 1)

    async def test_process_message_reject_by_validation_without_requeue(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], requeue_on_validation_fail=False)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1
            if handler_called < 2:
                raise ValidationException()

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(2)
        self.assertEquals(handler_called, 1)

    async def test_process_message_reject_by_validation_before_handler(self):
        handler_called = 0

        @self.app.consume_amqp(
            [self.input_queue], validation_schema=self.schema)
        async def handler(messages):
            nonlocal handler_called
            handler_called = handler_called + 1

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data={"wrong": "key"}
        )
        await asyncio.sleep(1)
        self.assertEquals(handler_called, 0)

    async def test_process_messages_and_requeue_only_one(self):
        first_read = set()
        second_read = set()

        @self.app.consume_amqp(
            [self.input_queue],
            coroutines=len(self.messages),
            bulk_flush_interval=0.1
        )
        async def handler(message):
            nonlocal first_read
            nonlocal second_read
            value = message.body["key"]
            if value not in first_read:
                first_read.add(value)
                if message.body == self.messages[0]:
                    # process only messages[0] again
                    raise Exception()
            else:
                second_read.add(value)

        await self.app.startup()
        await self.send_all_messages()

        await asyncio.sleep(1)

        for message in self.messages:
            self.assertTrue(message["key"] in first_read)

        self.assertSetEqual(second_read, {self.messages[0]["key"]})

    async def test_obtains_healthcheck(self):
        monitor = Monitor(Healthcheck(self.app))

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            pass

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)

        async with aiohttp.ClientSession() as session:
            timeout = aiohttp.ClientTimeout(total=1)
            url = f'http://{self.barterdude_host}:8080/healthcheck'
            async with session.get(url, timeout=timeout) as response:
                status_code = response.status
                text = await response.text()

        self.assertEquals(status_code, 200)
        self.assertEquals(
            text,
            '{"message": "Success rate: 1.0 (expected: 0.95)", '
            '"fail": 0, "success": 1, "status": "ok"}'
        )

    async def test_obtains_healthcheck_even_with_crashed_hook(self):
        monitor = Monitor(ErrorHook(), Healthcheck(self.app))

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            pass

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)

        async with aiohttp.ClientSession() as session:
            timeout = aiohttp.ClientTimeout(total=1)
            url = f'http://{self.barterdude_host}:8080/healthcheck'
            async with session.get(url, timeout=timeout) as response:
                status_code = response.status
                text = await response.text()

        self.assertEquals(status_code, 200)
        self.assertEquals(
            text,
            '{"message": "Success rate: 1.0 (expected: 0.95)", '
            '"fail": 0, "success": 1, "status": "ok"}'
        )

    async def test_obtains_prometheus_metrics(self):
        labels = {"app_name": "barterdude_consumer"}
        monitor = Monitor(Prometheus(self.app, labels))

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            pass

        await self.app.startup()
        await self.queue_manager.put(
            routing_key=self.input_queue,
            data=self.messages[0]
        )
        await asyncio.sleep(1)

        async with aiohttp.ClientSession() as session:
            timeout = aiohttp.ClientTimeout(total=1)
            url = f'http://{self.barterdude_host}:8080/metrics'
            async with session.get(url, timeout=timeout) as response:
                status_code = response.status
                text = await response.text()

        self.assertEquals(status_code, 200)
        self.assertNotEquals(-1, text.find(
            'barterdude_received_number_before_consume_messages_total'
            '{app_name="barterdude_consumer"} 1.0'))

    async def test_register_multiple_prometheus_hooks(self):
        """This test raised the following error:
           ValueError: Duplicated timeseries in CollectorRegistry"""

        Monitor(
            Prometheus(self.app, {"app_name": "barterdude1"}, "/metrics1"),
            Prometheus(self.app, {"app_name": "barterdude2"}, "/metrics2"),
        )

    async def test_print_logs(self):
        monitor = Monitor(Logging())
        error = Exception("raise expected")

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            if message.body == self.messages[0]:
                raise error

        await self.app.startup()

        with self.assertLogs("barterdude") as cm:
            await self.queue_manager.put(
                routing_key=self.input_queue,
                data=self.messages[0]
            )
            await asyncio.sleep(1)

        key = self.messages[0]["key"]
        error_str = repr(error)

        self.assertIn("'message': 'Before consume message'", cm.output[0])
        self.assertIn(
            f"'message_body': '{{\"key\": \"{key}\"}}'", cm.output[0]
        )
        self.assertIn("'delivery_tag': 1", cm.output[0])

        self.assertIn("'message': 'Failed to consume message'", cm.output[1])
        self.assertIn(
            f"'message_body': '{{\"key\": \"{key}\"}}'", cm.output[1]
        )
        self.assertIn("'delivery_tag': 1", cm.output[1])
        self.assertIn(f"'exception': \"{error_str}\"", cm.output[1])
        self.assertIn("'traceback': [", cm.output[1])

    async def test_fails_to_connect_to_rabbitmq(self):
        monitor = Monitor(Logging())

        self.app = BarterDude(hostname="invalid_host")

        @self.app.consume_amqp([self.input_queue], monitor)
        async def handler(message):
            pass

        await self.app.startup()
        with self.assertLogs("barterdude") as cm:
            await asyncio.sleep(2)

        self.assertIn(
            "{'message': 'Failed to connect to the broker', 'retries': 1,",
            cm.output[0]
        )



from traceback import format_tb
from typing import Iterable, Callable, Optional, Any

from asyncio import gather
from asyncworker.rabbitmq.message import RabbitMQMessage

from barterdude.conf import getLogger, BARTERDUDE_DEFAULT_LOG_LEVEL


class Monitor:
    def __init__(self, *hooks: Iterable):
        self.__hooks = hooks
        self._logger = getLogger(
            "hook.monitor", BARTERDUDE_DEFAULT_LOG_LEVEL)

    async def _callback(self,
                        method: Callable[[RabbitMQMessage], Optional[Any]],
                        message: RabbitMQMessage,
                        error: Optional[Exception] = None):
        try:
            return await (method(message, error) if error else method(message))
        except Exception as e:
            self._logger.error({
                "message": f"Error on hook method {method}",
                "exception": repr(e),
                "traceback": format_tb(e.__traceback__),
            })

    def _prepare_callbacks(self, method_name: str,
                           message: RabbitMQMessage,
                           error: Optional[Exception] = None):
        callbacks = []
        for hook in self.__hooks:
            callbacks.append(
                self._callback(getattr(hook, method_name), message, error)
            )
        return callbacks

    async def dispatch_before_consume(self, message: RabbitMQMessage):
        await gather(*self._prepare_callbacks("before_consume", message))

    async def dispatch_on_success(self, message: RabbitMQMessage):
        await gather(*self._prepare_callbacks("on_success", message))

    async def dispatch_on_fail(self, message: RabbitMQMessage,
                               error: Exception):
        await gather(*self._prepare_callbacks("on_fail", message, error))

    async def dispatch_on_connection_fail(
            self, error: Exception, retries: int
    ):
        await gather(*self._prepare_callbacks(
            "on_connection_fail", error, retries
        ))



from copy import copy
from prometheus_client import (
    CollectorRegistry,
    Counter,
    Histogram
)
from typing import Iterable
from barterdude.hooks.metrics.prometheus.metrics import Metrics


class Definitions:

    MESSAGE_UNITS = "messages"
    ERROR_UNITS = "errors"
    TIME_UNITS = "seconds"
    NAMESPACE = "barterdude"
    BEFORE_CONSUME = "before_consume"
    SUCCESS = "success"
    FAIL = "fail"
    TIME_MEASURE = "time_measure"
    CONNECTION_FAIL = "connection_fail"

    def __init__(
        self,
        registry: CollectorRegistry,
        metrics: Metrics,
        labelkeys: Iterable[str],
        time_buckets: tuple = Histogram.DEFAULT_BUCKETS
    ):
        self.__registry = registry
        self.__labelkeys = labelkeys
        self.__time_buckets = time_buckets
        self.__metrics = metrics

    def save_metrics(self):
        self._prepare_before_consume(
            self.BEFORE_CONSUME,
            namespace=self.NAMESPACE,
            unit=self.MESSAGE_UNITS,
        )
        self._prepare_on_complete(
            self.SUCCESS,
            namespace=self.NAMESPACE,
            unit=self.MESSAGE_UNITS,
        )
        self._prepare_on_complete(
            self.FAIL,
            namespace=self.NAMESPACE,
            unit=self.MESSAGE_UNITS,
        )
        self._prepare_time_measure(
            self.TIME_MEASURE,
            namespace=self.NAMESPACE,
            unit=self.TIME_UNITS,
        )
        self._prepare_on_connection_fail(
            self.CONNECTION_FAIL,
            namespace=self.NAMESPACE,
            unit=self.ERROR_UNITS,
        )

    def _prepare_before_consume(
            self, name: str, namespace: str = "", unit: str = ""):

        self.__metrics[name] = Counter(
                name="received_number_before_consume",
                documentation="Messages that worker received from queue(s)",
                labelnames=copy(self.__labelkeys),
                namespace=namespace,
                unit=unit,
                registry=self.__registry
            )

    def _prepare_on_complete(
            self, state: str, namespace: str = "", unit: str = ""):

        self.__metrics[state] = Counter(
                name=f"consumed_number_on_{state}",
                documentation=(f"Messages that worker consumed with {state}"
                               " from queue(s)"),
                labelnames=self.__labelkeys + ["state", "error"],
                namespace=namespace,
                unit=unit,
                registry=self.__registry
            )

    def _prepare_time_measure(
            self, name: str, namespace: str = "", unit: str = ""):

        self.__metrics[name] = Histogram(
                name="time_spent_processing_message",
                documentation=("Time spent when function was "
                               "processing a message"),
                buckets=self.__time_buckets,
                labelnames=self.__labelkeys + ["state", "error"],
                namespace=namespace,
                unit=unit,
                registry=self.__registry,
            )

    def _prepare_on_connection_fail(
            self, state: str, namespace: str, unit: str):

        self.__metrics[state] = Counter(
                name="connection_fail",
                documentation=("Number of times barterdude failed "
                               "to connect to the AMQP broker"),
                labelnames=self.__labelkeys,
                namespace=namespace,
                unit=unit,
                registry=self.__registry
            )
